from typing import Dict, Type, List

from waio.middleware import BaseMiddleware
from waio.rules.abc import ABCRule
from waio.rules.default import (
    MessageCommandsRule,
    StateRule,
    RegexRule,
    ContentTypeRule,
    TextRuleEquals,
    TextRuleContains,
    TextRuleStartswith,
    TextRuleEndswith,
    CallbackFilter,
)


class BotLabeler:
    DEFAULT_RULES: Dict[str, Type[ABCRule]] = {
        "commands": MessageCommandsRule,
        "state": StateRule,
        "regex": RegexRule,
        "text_equals": TextRuleEquals,
        "text_contains": TextRuleContains,
        "text_startswith": TextRuleStartswith,
        "text_endswith": TextRuleEndswith,
        "content_type": ContentTypeRule,
        "callback":  CallbackFilter,
    }
    MIDDLEWARES: List[BaseMiddleware] = []

    def __init__(self):
        self._custom_rules = {}

    def bind_rule(self, name: str, value: ABCRule):
        self._custom_rules[name] = value

    @property
    def custom_rules(self):
        return self._custom_rules

    @property
    def default_rules(self):
        return self.DEFAULT_RULES

    def register_middleware(self, middleware):
        self.MIDDLEWARES.append(middleware)



import aiohttp
import ujson

from urllib.parse import unquote
from typing import Optional, Union, Dict, Any, List, Tuple
from aiohttp import ContentTypeError

from waio.client.exceptions import FailedDecodeJson


class HTTPClient:
    async def request(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Tuple[Dict[str, Any], int]:
        if not headers:
            headers = {}

        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.request(method=method, url=url, **kwargs) as resp:
                return await self.generate_json_from_response(resp)

    async def generate_json_from_response(
        self,
        resp: aiohttp.ClientResponse
    ) -> Tuple[Dict[str, Any], int]:
        content_type = resp.headers.get('Content-Type')
        try:
            if content_type == 'text/plain':
                resp_text = await resp.text()
                resp_json = ujson.loads(resp_text)
                return self.decode_json(resp_json), resp.status

            elif content_type == 'application/json':
                resp_json = await resp.json()
                return resp_json, resp.status

        except ContentTypeError as e:
            raise FailedDecodeJson(f"Check args, URL is invalid - {e}")

    @staticmethod
    def decode_json(data: Union[List, Dict[str, Any]]):
        data_dumps = ujson.dumps(data, ensure_ascii=False)
        decoded_data_str = unquote(data_dumps)
        data_data_json = ujson.loads(decoded_data_str)
        return data_data_json

    @staticmethod
    def prepare_url(base_url: str, url: str, **kwargs):
        url = f"{base_url}/{url.format(**kwargs)}"
        return url



from typing import Optional

import ujson


class FileModel:
    def __init__(
            self,
            url: str,
            filename: str,
            caption: Optional[str] = None
    ):
        self.type = 'file'
        self.url = url
        self.filename = filename
        self.caption = caption

    def dict(self):
        return {
            "type": self.type,
            "url": self.url,
            "filename": self.filename,
            "caption": self.caption,
        }

    def json(self):
        return ujson.dumps(self.dict(), indent=2)



from typing import Optional
from typing import Protocol

from waio.keyboard.list import ListMessage
from waio.keyboard.reply import QuickReply


class Bot(Protocol):

    async def send_message(self, receiver: int, message: str):
        """Send message"""

    async def send_list(self, receiver: int, keyboard: ListMessage):
        """Send list message"""

    async def send_reply(self, receiver: int, keyboard: QuickReply):
        """Send reply message"""

    async def send_image(
            self,
            receiver: int,
            original_url: str,
            preview_url: Optional[str] = None,
            caption: Optional[str] = None,
    ):
        """Send image message"""

    async def send_file(
            self,
            receiver: int,
            url: str,
            filename: str,
            caption: Optional[str] = None,
    ):
        """Send file message"""

    async def send_video(
            self,
            receiver: int,
            url: str,
            caption: Optional[str] = None,
    ):
        """Send video message"""

    async def send_audio(
            self,
            receiver: int,
            url: str,
    ):
        """Send audio message"""

    async def send_sticker(
            self,
            receiver: int,
            url: str,
    ):
        """Send sticker message"""



from typing import Union, Dict, List


def clear_none_values(d: Union[List, Dict]):
    if d is None:
        return None
    elif isinstance(d, list):
        return list(filter(lambda x: x is not None, map(clear_none_values, d)))
    elif not isinstance(d, dict):
        return d
    else:
        r = dict(
            filter(lambda x: x[1] is not None,
                   map(lambda x: (x[0], clear_none_values(x[1])),
                       d.items())))
        if not bool(r):
            return None
        return r



from typing import Optional

import ujson


class ImageModel:
    def __init__(
        self,
        original_url: str,
        preview_url: Optional[str] = None,
        caption: Optional[str] = None
    ):
        self.type = 'image'
        self.original_url = original_url
        self.preview_url = preview_url or original_url
        self.caption = caption

    def dict(self):
        return {
            "type": self.type,
            "originalUrl": self.original_url,
            "previewUrl": self.preview_url,
            "caption": self.caption,
        }

    def json(self):
        return ujson.dumps(self.dict(), indent=2)



import ujson


class MessageText:
    def __init__(self, text: str):
        self.type = 'text'
        self.text = text

    def dict(self):
        return {"type": self.type, "text": self.text}

    def json(self):
        return ujson.dumps(self.dict(), indent=2)



from typing import Union, Optional, Dict, Any

from waio.states.fsm import BaseState
from waio.storage.redis import RedisStorage


class FSMContext:
    def __init__(self, storage: RedisStorage, user_phone: int):
        self.storage = storage
        self.user = user_phone

    async def set_state(self, value: BaseState) -> bool:
        return await self.storage.set_state(key=self.user, value=value)

    async def get_state(self) -> str:
        return await self.storage.get_state(key=self.user)

    async def finish(self, clear_data: Optional[bool] = False) -> None:
        await self.storage.finish_state(key=self.user)
        if clear_data:
            await self.storage.finish_data(key=self.user)

    async def set_data(self, **kwargs: Union[str, int]) -> bool:
        return await self.storage.set_data(self.user, **kwargs)

    async def get_data(self, *data_keys) -> Dict[str, Union[str, int]]:
        data_values = await self.storage.get_data(key=self.user)
        if data_keys:
            return {key: value for key, value in data_values.items() if key in data_keys}
        return data_values



import ujson
from typing import (
    Optional,
    Any, Dict, Union
)

from waio.storage.connector import RedisConnector


class RedisStorage(RedisConnector):
    def __init__(self, redis_url: str, prefix_fsm: str = 'fsm', db: int = 1):
        super().__init__(redis_url=redis_url, db_number=db)
        self.prefix_fsm = prefix_fsm

    def _generate_key_state(self, key) -> str:
        return f"{self.prefix_fsm}:{key}:state"

    def _generate_key_data(self, key) -> str:
        return f"{self.prefix_fsm}:{key}:data"

    async def _set(self, key: str, value: Any) -> bool:
        _value = ujson.dumps(value)
        set_response = await self.redis.set(name=key, value=_value)
        return bool(set_response)

    async def _key_availability(self, key) -> bool:
        key_exist = await self.redis.exists(key)
        if key_exist:
            return True
        return False

    async def _base_get(self, key) -> Dict[Union[str, int], Any]:
        value = await self.redis.get(name=key)
        value_dict = ujson.loads(value)
        return value_dict

    async def set(self, key, value) -> bool:
        if await self._key_availability(key=key):
            key_value = await self._base_get(key=key)
            new_values = {**key_value, **value}
            return await self._set(key, new_values)
        else:
            return await self._set(key, value)

    async def set_state(self, key, value) -> bool:
        fsm_key_state = self._generate_key_state(key=key)
        state_dict = {"state": str(value)}
        return await self.set(key=fsm_key_state, value=state_dict)

    async def get_state(self, key) -> Optional[str]:
        key_state = self._generate_key_state(key=key)
        query = await self.redis.get(name=key_state)
        if query:
            query_state = ujson.loads(query)
            if query_state.get("state"):
                return query_state["state"]
        return None

    async def finish_state(self, key) -> None:
        key_state = self._generate_key_state(key=key)
        await self.redis.delete(key_state)

    async def set_data(self, key, **kwargs) -> bool:
        fsm_key_data = self._generate_key_data(key=key)
        return await self.set(key=fsm_key_data, value=kwargs)

    async def get_data(self, key):
        key = self._generate_key_data(key=key)
        return await self._base_get(key)

    async def finish_data(self, key) -> None:
        key_data = self._generate_key_data(key=key)
        await self.redis.delete(key_data)



import json

from waio.utils.form_data import CustomFormData


def generate_message_form(source, receiver, app_name, message: json) -> CustomFormData:
    form = CustomFormData()
    form.add_www_form(name='channel', value='whatsapp')
    form.add_www_form(name='source', value=source)
    form.add_www_form(name='destination', value=receiver)
    form.add_www_form(name='src.name', value=app_name)
    form.add_www_form(name='message', value=message)

    return form



import re
from typing import List, Union, Optional, Dict, Literal

from waio.rules.abc import ABCRule
from waio.states import BaseState
from waio.types.content_types import ContentType
from waio.types.message import Message
from waio.utils.callback.filters import CallbackDataFilterItem


class MessageCommandsRule(ABCRule):
    def __init__(self, commands: List[str], prefix: str = "/"):
        self.prefix = prefix
        self.commands = commands

    async def check(self, message: Message) -> bool:
        for command in self.commands:
            command_and_prefix = f"{self.prefix}{command}"
            if message.text.startswith(self.prefix):
                if command_and_prefix == message.text:
                    return True
        return False


class StateRule(ABCRule):
    def __init__(self, state: Optional[Union[BaseState, Literal['*']]] = None):
        self.state = state

    async def check(self, message: Message) -> Union[dict, bool]:
        user_state = await message.current_state()
        if str(self.state) == user_state or not self.state or self.state == '*':
            return True
        return False


class RegexRule(ABCRule):
    def __init__(self, regexp: Union[str, List[str], re.Pattern, List[re.Pattern]]):
        if isinstance(regexp, re.Pattern):
            regexp = [regexp]
        elif isinstance(regexp, str):
            regexp = [re.compile(regexp)]
        elif isinstance(regexp, list):
            regexp = [re.compile(exp) for exp in regexp]

        self.regexp = regexp

    async def check(self, message: Message) -> Union[Dict[str, re.Match], bool]:
        for regexp in self.regexp:
            match = re.match(regexp, message.text)
            if match:
                return {"regex": match}
        return False


class TextRuleEquals(ABCRule):
    def __init__(self, equals: List[str]):
        self.equals = equals

    async def check(self, message: Message) -> bool:
        for elem in self.equals:
            if elem == message.text:
                return True
        return False


class TextRuleContains(ABCRule):
    def __init__(self, contains: List[str]):
        self.contains = contains

    async def check(self, message: Message) -> bool:
        for elem in self.contains:
            if elem in message.text:
                return True
        return False


class TextRuleStartswith(ABCRule):
    def __init__(self, startswith: List[str]):
        self.startswith = startswith

    async def check(self, message: Message) -> bool:
        for elem in self.startswith:
            if message.text.startswith(elem):
                return True
        return False


class TextRuleEndswith(ABCRule):
    def __init__(self, endswith: List[str]):
        self.endswith = endswith

    async def check(self, message: Message) -> bool:
        for elem in self.endswith:
            if message.text.endswith(elem):
                return True
        return False


class TextRule(ABCRule):
    def __init__(
        self,
        equals: Optional[List[str]] = None,
        contains: Optional[List[str]] = None,
        startswith: Optional[List[str]] = None,
        endswith: Optional[List[str]] = None
    ):
        self.equals = equals
        self.contains = contains
        self.startswith = startswith
        self.endswith = endswith

    async def check(self, message: Message) -> bool:
        if self.equals:
            eq = TextRuleEquals(equals=self.equals)
            return await eq.check(message)

        if self.contains:
            ct = TextRuleContains(contains=self.equals)
            return await ct.check(message)

        if self.startswith:
            ss = TextRuleStartswith(startswith=self.startswith)
            return await ss.check(message)

        if self.endswith:
            es = TextRuleEndswith(endswith=self.endswith)
            return await es.check(message)


class ContentTypeRule(ABCRule):
    def __init__(self, content_types: List[ContentType]):
        self.content_types = content_types

    async def check(self, message: Message) -> Union[Dict[str, re.Match], bool]:
        for content_type in self.content_types:
            if isinstance(message.message.payload, content_type.value):
                return True
        return False


class PhoneNumberRule(ABCRule):
    def __init__(self, phones: List[int]):
        self.phones = phones

    async def check(self, message: Message) -> bool:
        for phone in self.phones:
            if phone == message.message.payload.sender.phone:
                return True
        return False


class CallbackFilter(ABCRule):
    def __init__(self, item: CallbackDataFilterItem):
        self.item = item

    async def check(self, message: Message):
        return await self.item.check(message=message)



from typing import Callable, Union, Any
from waio.rules.abc import ABCRule


class FromFuncHandler:
    def __init__(self, handler: Callable, *rules: ABCRule):
        self.handler = handler
        self.rules = rules

    async def filter(self, event: Any) -> Union[dict, bool]:
        rule_context = {}
        for rule in self.rules:
            result = await rule.check(event)
            if result is False or result is None:
                return False
            elif result is True:
                continue
            rule_context.update(result)
        return rule_context

    async def handle(self, **context) -> Any:
        return await self.handler(**context)

    def __repr__(self):
        return (
            f"<FromFuncHandler {self.handler.__name__} "
            f"rules={self.rules}>"
        )



from dataclass_factory import Schema
from dataclass_factory.schema_helpers import type_checker


PAYLOAD_TEXT = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
        "text": ("payload", "text"),
    },
    skip_internal=True
)

PAYLOAD_SENDER = Schema(
    name_mapping={
        "phone": "phone",
        "name": "name",
    },
    skip_internal=True
)

PAYLOAD_LIST = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
        "title": ("payload", "title"),
        "id": ("payload", "id"),
        "reply": ("payload", "reply"),
        "postback_text": ("payload", "postbackText"),
        "description": ("payload", "description"),
    },
    pre_parse=type_checker("list_reply", field="type"),
    skip_internal=True
)

PAYLOAD_REPLY_KEYBOARD = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
        "title": ("payload", "title"),
        "id": ("payload", "id"),
        "reply": ("payload", "reply"),
    },
    pre_parse=type_checker(value="button_reply", field="type"),
    skip_internal=True
)

PAYLOAD_IMAGE = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
        "url": ("payload", "url"),
        "content_type": ("payload", "contentType"),
        "url_expiry": ("payload", "urlExpiry"),
        "caption": ("payload", "caption")
    },
    pre_parse=type_checker(value="image", field="type"),
    skip_internal=True
)

PAYLOAD_FILE = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
        "url": ("payload", "url"),
        "content_type": ("payload", "contentType"),
        "url_expiry": ("payload", "urlExpiry"),
        "caption": ("payload", "caption"),
        "name": ("payload", "name"),
    },
    pre_parse=type_checker(value="file", field="type"),
    skip_internal=True
)

BASE_MODEL = Schema(
    name_mapping={
        "payload_id": "id",
        "type": "type",
    },
    skip_internal=True
)

CONTACTS_PAYLOAD = Schema(
    name_mapping={
        "payload_id": "id",
        "contacts": ("payload", "contacts"),
    },
    skip_internal=True
)

PAYLOAD_CONTACT_PHONE = Schema(
    name_mapping={
        "phone": "phone",
        "type": "type",
    },
    skip_internal=True
)

PAYLOAD_CONTACT_ORGANIZATION = Schema(
    name_mapping={
        "company": "company",
    },
    skip_internal=True
)

PAYLOAD_CONTACT_URL = Schema(
    name_mapping={
        "url": "url",
        "type": "type"
    },
    skip_internal=True
)

PAYLOAD_CONTACT_INFORMATION = Schema(
    name_mapping={
        "service": "service",
        "user_id": "user_id"
    },
    skip_internal=True
)

PAYLOAD_CONTACT = Schema(
    name_mapping={
        "addresses": "addresses",
        "emails": "emails",
        "ims": "ims",
        "name": "name",
        "org": "org",
        "phones": "phones",
        "urls": "urls",
    },
    skip_internal=True
)

PAYLOAD_CONTACT_EMAIL = Schema(
    name_mapping={
        "email": "email",
        "type": "type"
    },
    skip_internal=True
)
PAYLOAD_CONTACT_ADDRESS = Schema(
    name_mapping={
        "city": "city",
        "country": "country",
        "country_code": "countryCode",
        "state": "state",
        "street": "street",
        "type": "type",
        "zip": "zip"
    },
    skip_internal=True
)

RESPONSE_MESSAGE_PAYLOAD_PAYLOAD = Schema(
    name_mapping={
        "whatsapp_Message_id": "whatsappMessageId",
        "type": "type",
    },
    skip_internal=True
)

RESPONSE_MESSAGE_PAYLOAD = Schema(
    name_mapping={
        "id": "id",
        "type": "type",
        "destination": "destination",
        "payload": "payload"
    },
    skip_internal=True
)

RESPONSE_MESSAGE = Schema(
    name_mapping={
        "app": "app",
        "timestamp": "timestamp",
        "version": "version",
        "type": "type",
        "payload": "payload"
    },
    skip_internal=True
)



from typing import Any, Dict, Optional

from waio.client.http import HTTPClient
from waio.factory.base import ResponseModel
from waio.factory.factory import factory_gupshup
from waio.gupshup.api import GupshupSettings
from waio.gupshup.form import generate_message_form
from waio.handlers.base_handlers import Handler, BaseHandlers
from waio.handlers.executor import HandlerExecutor
from waio.keyboard.list import ListMessage
from waio.keyboard.reply import QuickReply
from waio.labeler import BotLabeler
from waio.logs.logger import logger
from waio.middleware import MiddlewareResponse
from waio.models.audio import AudioModel
from waio.models.enums import GupshupMethods
from waio.models.file import FileModel
from waio.models.image import ImageModel
from waio.models.sticker import StickerModel
from waio.models.text import MessageText
from waio.models.video import VideoModel
from waio.states.context import FSMContext
from waio.storage.redis import RedisStorage
from waio.types.message import Message


class Bot(GupshupSettings, HTTPClient):

    def __init__(self, apikey: str, src_name: str, phone_number: int):
        super().__init__(apikey=apikey, src_name=src_name, phone_number=phone_number)

    async def _base_request(self, receiver: int, data: Any):
        form = self._generate_form(receiver=receiver, data=data)
        response = await self.request(
            headers=self._headers(),
            method='POST',
            url=GupshupMethods.message.value,
            data=form,
        )
        return response

    def _generate_form(self, receiver, data):
        return generate_message_form(
            source=self.phone_number,
            receiver=receiver,
            app_name=self.src_name,
            message=data.json()
        )

    async def send_message(self, receiver: int, message: str):
        msg = MessageText(text=message)
        return await self._base_request(receiver=receiver, data=msg)

    async def send_image(
            self,
            receiver: int,
            original_url: str,
            preview_url: Optional[str] = None,
            caption: Optional[str] = None,
    ):
        image = ImageModel(original_url=original_url, preview_url=preview_url, caption=caption)
        return await self._base_request(receiver=receiver, data=image)

    async def send_file(
            self,
            receiver: int,
            url: str,
            filename: str,
            caption: Optional[str] = None,
    ):
        file = FileModel(url=url, filename=filename, caption=caption)
        return await self._base_request(receiver=receiver, data=file)

    async def send_video(
            self,
            receiver: int,
            url: str,
            caption: Optional[str] = None,
    ):
        video = VideoModel(url=url, caption=caption)
        return await self._base_request(receiver=receiver, data=video)

    async def send_audio(
            self,
            receiver: int,
            url: str,
    ):
        audio = AudioModel(url=url)
        return await self._base_request(receiver=receiver, data=audio)

    async def send_sticker(
            self,
            receiver: int,
            url: str,
    ):
        sticker = StickerModel(url=url)
        return await self._base_request(receiver=receiver, data=sticker)

    async def send_list(self, receiver: int, keyboard: ListMessage):
        return await self._base_request(receiver=receiver, data=keyboard)

    async def send_reply(self, receiver: int, keyboard: QuickReply):
        return await self._base_request(receiver=receiver, data=keyboard)


class Dispatcher(Handler, BaseHandlers):
    def __init__(self, bot: Bot, storage: Optional[RedisStorage] = None):
        self.bot = bot
        self.storage = storage
        self.labeler = BotLabeler()
        super().__init__(self.labeler)

    def state(self, user_phone: int) -> FSMContext:
        return FSMContext(storage=self.storage, user_phone=user_phone)

    async def handle_event(self, event: Dict[str, Any]) -> None:
        try:
            logger.debug(f'Event: - {event}')
            if event.get("type") == "message":
                context_variables = {}

                data_load = factory_gupshup.load(event, ResponseModel)
                message = Message(bot=self.bot, message=data_load, state_func=self.state)

                for middleware in self.labeler.MIDDLEWARES:
                    middleware.fill(event=message)
                    response = await middleware.pre()

                    logger.debug(f'[PRE]-Middleware: {middleware}')

                    if response == MiddlewareResponse(False):
                        return
                    elif isinstance(response, dict):
                        context_variables.update(response)

                logger.debug(f'[PRE]-Middleware values: {context_variables}')

                handle_responses = []
                current_handlers = []

                for handler in self.handlers:
                    resp = await HandlerExecutor.execute(handler=handler, message=message, **context_variables)

                    if resp:
                        handle_responses.append(resp)
                        current_handlers.append(handler)
                        await handler.handle(**resp)
                        break

                logger.debug(f'Handlers: {current_handlers}, Return: {handle_responses}')

                for middleware in reversed(self.labeler.MIDDLEWARES):
                    middleware.fill(event=message, handlers=current_handlers, handle_responses=handle_responses)
                    logger.debug(f'[POST]-Middleware - {middleware}')

                    await middleware.post()
        except Exception as exc:
            logger.exception('Exception', exc)




from typing import List, Callable

from waio.handlers import ABCHandler, FromFuncHandler
from waio.labeler import BotLabeler


class BaseHandlers:

    def __init__(self, labeler: BotLabeler):
        self.handlers: List[ABCHandler] = []
        self.labeler = labeler

    def add_message_handler(self, handler):
        self.handlers.append(handler)

    def register_message_handler(
        self,
        handler: Callable,
        *rules,
        **custom_rules,
    ):
        handler_object = FromFuncHandler(
            handler,
            *rules,
            *self.base_rules(**custom_rules),
            *self.custom_rules(**custom_rules)
        )
        self.add_message_handler(handler=handler_object)

    def base_rules(self, **rules):
        default_rules = [
            self.labeler.default_rules[k](v)
            for k, v in rules.items()
            if k in self.labeler.default_rules.keys()
        ]

        return default_rules

    def custom_rules(self, **rules):
        custom = [
            self.labeler.custom_rules[k](v)
            for k, v in rules.items()
            if k in self.labeler.custom_rules.keys()
        ]

        return custom


class Handler(BaseHandlers):

    def message_handler(
        self,
        *rules,
        **custom_rules
    ):
        def decorator(handler) -> None:

            handler_object = FromFuncHandler(
                handler,
                *rules,
                *self.base_rules(**custom_rules),
                *self.custom_rules(**custom_rules)
            )
            self.add_message_handler(handler=handler_object)

            return handler
        return decorator



import ujson


class AudioModel:
    def __init__(
            self,
            url: str,
    ):
        self.type = 'audio'
        self.url = url

    def dict(self):
        return {
            "type": self.type,
            "url": self.url,
        }

    def json(self):
        return ujson.dumps(self.dict(), indent=2)



import ujson
from typing import Union, List, Optional
from waio.keyboard.list import ListMainButton


class KeyboardButton(ListMainButton):
    def __init__(self, title: str):
        super().__init__(title)


class QuickReplyContentBase:
    def __init__(self, text: str, caption: str):
        self.text = text
        self.caption = caption


class QuickReplyContentText(QuickReplyContentBase):
    def __init__(self, header: str, text: str, caption: str):
        self.type = "text"
        self.header = header

        super().__init__(text, caption)

    def json(self):
        return {
            "type": self.type,
            "header": self.header,
            "text": self.text,
            "caption": self.caption
        }


class QuickReplyContentImage(QuickReplyContentBase):
    def __init__(self, url: str, text: str, caption: str):
        self.type = "image"
        self.url = url

        super().__init__(text, caption)

    def json(self):
        return {
            "type": self.type,
            "url": self.url,
            "text": self.text,
            "caption": self.caption
        }


class QuickReplyContentDocument(QuickReplyContentBase):
    def __init__(self, url: str, filename: str, text: str, caption: str):
        self.type = "file"
        self.url = url
        self.filename = filename

        super().__init__(text, caption)

    def json(self):
        return {
            "type": self.type,
            "url": self.url,
            "text": self.text,
            "caption": self.caption,
            "filename": self.filename
        }


class QuickReplyContentVideo(QuickReplyContentBase):
    def __init__(self, url: str, text: str, caption: str):
        self.type = "video"
        self.url = url

        super().__init__(text, caption)

    def json(self):
        return {
            "type": self.type,
            "url": self.url,
            "text": self.text,
            "caption": self.caption,
        }


class QuickReply:
    def __init__(
            self,
            callback_data: str,
            content: Union[
                QuickReplyContentText,
                QuickReplyContentImage,
                QuickReplyContentDocument,
                QuickReplyContentVideo
            ],
            options: Optional[List[KeyboardButton]] = None
    ):
        self.type = "quick_reply"
        self.callback_data = callback_data
        self.content = content

        if options is None:
            self.options = []
        else:
            self.options = options

    def add(self, element: KeyboardButton) -> 'QuickReply':
        self.options.append(element)
        return self

    def dict(self):
        return {
            "type": self.type,
            "msgid": self.callback_data,
            "content": self.content.json(),
            "options": [element.json() for element in self.options]
        }

    def json(self):
        return ujson.dumps(self.dict(), indent=2)



from enum import Enum


class RequestMethods(str, Enum):
    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class ResponseGupshupMessageType(str, Enum):
    request = 'message'
    message_event = 'message-event'
    event = 'user-event'


class ResponseGupshupPayloadType(str, Enum):
    enqueued = 'enqueued'
    text = 'text'
    file = 'file'
    contact = 'contact'
    image = 'image'


class EventPayloadType(str, Enum):
    start_dialog = 'opted-in'
    session = 'session'
    sandbox_start = 'SANDBOX_START'
    failed = 'failed'


class GupshupMethods(str, Enum):
    message = 'https://api.gupshup.io/sm/api/v1/msg'




from typing import Optional, Dict, Any, Union


class BaseState:
    def __init__(
        self,
        state: Optional[str] = None,
        group: Optional[object] = None
    ):
        self._group = group
        self._state = state

    def __repr__(self):
        return f"<BaseState> {self.group.__name__}:{self.state}"

    def __str__(self):
        return f"{self.group.__name__}:{self.state}"

    def __set_name__(self, owner: "StatesGroup", name: str) -> None:
        self.group = owner
        self.state = name

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def group(self):
        return self._group

    @group.setter
    def group(self, value):
        self._group = value


class StatesGroupMeta(type):
    __states__: Dict[str, Any]

    def __new__(mcs, cls_name, bases, attrs: Dict[str, Union[Any, BaseState]]):
        mcs.__states__ = {
            state_attr: state_value
            for state_attr, state_value in attrs.items()
            if not state_attr.startswith("__")
        }

        return super(StatesGroupMeta, mcs).__new__(
            mcs, cls_name, bases, attrs)


class StatesGroup(metaclass=StatesGroupMeta):
    pass


class State(BaseState):
    pass



from dataclasses import dataclass
from uuid import uuid4

from waio.models.enums import ResponseGupshupMessageType, ResponseGupshupPayloadType, EventPayloadType


@dataclass
class ResponseMessagePayloadPayload:
    whatsapp_Message_id: str
    type: EventPayloadType


@dataclass
class ResponseMessagePayload:
    id: uuid4
    type: ResponseGupshupPayloadType
    destination: str
    payload: ResponseMessagePayloadPayload


@dataclass
class ResponseMessage:
    app: str
    timestamp: int
    version: int
    type: ResponseGupshupMessageType
    payload: ResponseMessagePayload





class GupshupSettings:
    def __init__(self, apikey: str, src_name: str, phone_number: int):
        self.apikey = apikey
        self.src_name = src_name
        self.phone_number = phone_number

    def _headers(self):
        headers = {
            'Cache-Control': 'no-cache',
            'cache-control': 'no-cache',
            'Content-Type': 'application/x-www-form-urlencoded',
            'apikey': self.apikey
        }
        return headers



from dataclasses import dataclass
from typing import Union, Optional

from waio.factory.models.basic import PayloadBaseModel


@dataclass
class PayloadImage(PayloadBaseModel):
    url: str
    content_type: str
    url_expiry: Union[int, str]
    caption: Optional[str] = None
